import { Time } from '../../../../domain/entities/ecosystem/time';

const MARGIN_DURATION = 900;
const VALIDITY_DURATION = 365;
const mockTimeConverter = {
  toUtcDate: jest.fn(),
};

describe('Time entity', () => {
  it('should be defined', () => {
    const time = new Time(
      {
        departure: '2023-04-01 12:24:00',
      },
      MARGIN_DURATION,
      VALIDITY_DURATION,
      mockTimeConverter,
    );
    expect(time).toBeDefined();
  });

  describe('init', () => {
    it('should initialize a punctual time request', () => {
      const time = new Time(
        {
          departure: '2023-04-01 12:24:00',
        },
        MARGIN_DURATION,
        VALIDITY_DURATION,
        mockTimeConverter,
      );
      time.init();
      expect(time.fromDate.getFullYear()).toBe(
        new Date('2023-04-01 12:24:00').getFullYear(),
      );
    });
    it('should initialize a punctual time request with specific single margin duration', () => {
      const time = new Time(
        {
          departure: '2023-04-01 12:24:00',
          marginDuration: 300,
        },
        MARGIN_DURATION,
        VALIDITY_DURATION,
        mockTimeConverter,
      );
      time.init();
      expect(time.marginDurations['tue']).toBe(300);
    });
    it('should initialize a punctual time request with specific margin durations', () => {
      const time = new Time(
        {
          departure: '2023-04-01 12:24:00',
          marginDurations: {
            sat: 350,
          },
        },
        MARGIN_DURATION,
        VALIDITY_DURATION,
        mockTimeConverter,
      );
      time.init();
      expect(time.marginDurations['tue']).toBe(900);
      expect(time.marginDurations['sat']).toBe(350);
    });
    it('should initialize a punctual time request with specific single margin duration and margin durations', () => {
      const time = new Time(
        {
          departure: '2023-04-01 12:24:00',
          marginDuration: 500,
          marginDurations: {
            sat: 350,
          },
        },
        MARGIN_DURATION,
        VALIDITY_DURATION,
        mockTimeConverter,
      );
      time.init();
      expect(time.marginDurations['tue']).toBe(500);
      expect(time.marginDurations['sat']).toBe(350);
    });
    it('should initialize a recurrent time request', () => {
      const time = new Time(
        {
          fromDate: '2023-04-01',
          schedule: {
            mon: '12:00',
          },
        },
        MARGIN_DURATION,
        VALIDITY_DURATION,
        mockTimeConverter,
      );
      time.init();
      expect(time.fromDate.getFullYear()).toBe(
        new Date('2023-04-01').getFullYear(),
      );
    });
    it('should throw an exception if no date is provided', () => {
      const time = new Time(
        {},
        MARGIN_DURATION,
        VALIDITY_DURATION,
        mockTimeConverter,
      );
      expect(() => time.init()).toThrow();
    });
    it('should throw an exception if punctual date is invalid', () => {
      const time = new Time(
        {
          departure: '2023-15-01 12:24:00',
        },
        MARGIN_DURATION,
        VALIDITY_DURATION,
        mockTimeConverter,
      );
      expect(() => time.init()).toThrow();
    });
    it('should throw an exception if recurrent fromDate is invalid', () => {
      const time = new Time(
        {
          fromDate: '2023-15-01',
        },
        MARGIN_DURATION,
        VALIDITY_DURATION,
        mockTimeConverter,
      );
      expect(() => time.init()).toThrow();
    });
    it('should throw an exception if recurrent toDate is invalid', () => {
      const time = new Time(
        {
          fromDate: '2023-04-01',
          toDate: '2023-13-01',
        },
        MARGIN_DURATION,
        VALIDITY_DURATION,
        mockTimeConverter,
      );
      expect(() => time.init()).toThrow();
    });
    it('should throw an exception if recurrent toDate is before fromDate', () => {
      const time = new Time(
        {
          fromDate: '2023-04-01',
          toDate: '2023-03-01',
        },
        MARGIN_DURATION,
        VALIDITY_DURATION,
        mockTimeConverter,
      );
      expect(() => time.init()).toThrow();
    });
    it('should throw an exception if schedule is missing', () => {
      const time = new Time(
        {
          fromDate: '2023-04-01',
          toDate: '2024-03-31',
        },
        MARGIN_DURATION,
        VALIDITY_DURATION,
        mockTimeConverter,
      );
      expect(() => time.init()).toThrow();
    });
    it('should throw an exception if schedule is empty', () => {
      const time = new Time(
        {
          fromDate: '2023-04-01',
          toDate: '2024-03-31',
          schedule: {},
        },
        MARGIN_DURATION,
        VALIDITY_DURATION,
        mockTimeConverter,
      );
      expect(() => time.init()).toThrow();
    });
    it('should throw an exception if schedule is invalid', () => {
      const time = new Time(
        {
          fromDate: '2023-04-01',
          toDate: '2024-03-31',
          schedule: {
            mon: '32:78',
          },
        },
        MARGIN_DURATION,
        VALIDITY_DURATION,
        mockTimeConverter,
      );
      expect(() => time.init()).toThrow();
    });
  });
  it('should throw an exception if margin durations is provided but empty', () => {
    const time = new Time(
      {
        departure: '2023-04-01 12:24:00',
        marginDurations: {},
      },
      MARGIN_DURATION,
      VALIDITY_DURATION,
      mockTimeConverter,
    );
    expect(() => time.init()).toThrow();
  });
});
