import { Test, TestingModule } from '@nestjs/testing';
import { MatchUseCase } from '../../../domain/usecases/match.usecase';
import { MatchRequest } from '../../../domain/dtos/match.request';
import { MatchQuery } from '../../../queries/match.query';
import { AutomapperModule } from '@automapper/nestjs';
import { classes } from '@automapper/classes';
import { IDefaultParams } from '../../../domain/types/default-params.type';
import { AlgorithmType } from '../../../domain/types/algorithm.enum';
import { Matcher } from '../../../domain/entities/engine/matcher';
import { Match } from '../../../domain/entities/ecosystem/match';
import {
  MatcherException,
  MatcherExceptionCode,
} from '../../../exceptions/matcher.exception';
import { MESSAGE_PUBLISHER } from '../../../../../app.constants';

const mockMatcher = {
  match: jest
    .fn()
    .mockImplementationOnce(() => [new Match(), new Match(), new Match()])
    .mockImplementationOnce(() => {
      throw new MatcherException(
        MatcherExceptionCode.INTERNAL,
        'Something terrible happened !',
      );
    }),
};

const mockMessagePublisher = {
  publish: jest.fn().mockImplementation(),
};

const mockGeorouterCreator = {
  create: jest.fn().mockImplementation(),
};

const mockTimezoneFinder = {
  timezones: jest.fn().mockImplementation(() => ['Europe/Paris']),
};

const mockTimeConverter = {
  toUtcDate: jest.fn(),
};

const defaultParams: IDefaultParams = {
  DEFAULT_UUID: '00000000-0000-0000-0000-000000000000',
  MARGIN_DURATION: 900,
  VALIDITY_DURATION: 365,
  DEFAULT_TIMEZONE: 'Europe/Paris',
  DEFAULT_SEATS: 3,
  DEFAULT_ALGORITHM_SETTINGS: {
    ALGORITHM: AlgorithmType.CLASSIC,
    STRICT: false,
    REMOTENESS: 15000,
    USE_PROPORTION: true,
    PROPORTION: 0.3,
    USE_AZIMUTH: true,
    AZIMUTH_MARGIN: 10,
    MAX_DETOUR_DISTANCE_RATIO: 0.3,
    MAX_DETOUR_DURATION_RATIO: 0.3,
    GEOROUTER_TYPE: 'graphhopper',
    GEOROUTER_URL: 'http://localhost',
  },
};

const matchRequest: MatchRequest = new MatchRequest();
matchRequest.waypoints = [
  {
    lon: 1.093912,
    lat: 49.440041,
  },
  {
    lat: 50.630992,
    lon: 3.045432,
  },
];
matchRequest.departure = '2023-04-01 12:23:00';

describe('MatchUseCase', () => {
  let matchUseCase: MatchUseCase;

  beforeAll(async () => {
    const module: TestingModule = await Test.createTestingModule({
      imports: [AutomapperModule.forRoot({ strategyInitializer: classes() })],
      providers: [
        {
          provide: MESSAGE_PUBLISHER,
          useValue: mockMessagePublisher,
        },
        {
          provide: Matcher,
          useValue: mockMatcher,
        },
        MatchUseCase,
      ],
    }).compile();

    matchUseCase = module.get<MatchUseCase>(MatchUseCase);
  });

  it('should be defined', () => {
    expect(matchUseCase).toBeDefined();
  });

  describe('execute', () => {
    it('should return matches', async () => {
      const matches = await matchUseCase.execute(
        new MatchQuery(
          matchRequest,
          defaultParams,
          mockGeorouterCreator,
          mockTimezoneFinder,
          mockTimeConverter,
        ),
      );
      expect(matches.total).toBe(3);
    });

    it('should throw an exception when error occurs', async () => {
      await expect(
        matchUseCase.execute(
          new MatchQuery(
            matchRequest,
            defaultParams,
            mockGeorouterCreator,
            mockTimezoneFinder,
            mockTimeConverter,
          ),
        ),
      ).rejects.toBeInstanceOf(MatcherException);
    });
  });
});
