import { Test, TestingModule } from '@nestjs/testing';
import { AD_REPOSITORY } from '@modules/ad/ad.di-tokens';
import { AggregateID } from '@mobicoop/ddd-library';
import { AdEntity } from '@modules/ad/core/domain/ad.entity';
import { ConflictException } from '@mobicoop/ddd-library';
import { CreateAdProps, Frequency } from '@modules/ad/core/domain/ad.types';
import { CreateAdService } from '@modules/ad/core/application/commands/create-ad/create-ad.service';
import { CreateAdCommand } from '@modules/ad/core/application/commands/create-ad/create-ad.command';
import { AdAlreadyExistsException } from '@modules/ad/core/domain/ad.errors';
import { WaypointProps } from '@modules/ad/core/domain/value-objects/waypoint.value-object';

const originWaypoint: WaypointProps = {
  position: 0,
  lon: 48.689445,
  lat: 6.17651,
};
const destinationWaypoint: WaypointProps = {
  position: 1,
  lon: 48.8566,
  lat: 2.3522,
};
const createAdProps: CreateAdProps = {
  id: '4eb6a6af-ecfd-41c3-9118-473a507014d4',
  fromDate: '2023-12-21',
  toDate: '2023-12-21',
  schedule: [
    {
      day: 4,
      time: '08:15',
      margin: 900,
    },
  ],
  driver: true,
  passenger: true,
  seatsProposed: 3,
  seatsRequested: 1,
  strict: false,
  frequency: Frequency.PUNCTUAL,
  waypoints: [originWaypoint, destinationWaypoint],
};

const mockAdRepository = {
  insert: jest
    .fn()
    .mockImplementationOnce(() => ({}))
    .mockImplementationOnce(() => {
      throw new Error();
    })
    .mockImplementationOnce(() => {
      throw new ConflictException('already exists');
    }),
};

describe('create-ad.service', () => {
  let createAdService: CreateAdService;

  beforeAll(async () => {
    const module: TestingModule = await Test.createTestingModule({
      providers: [
        {
          provide: AD_REPOSITORY,
          useValue: mockAdRepository,
        },
        CreateAdService,
      ],
    }).compile();

    createAdService = module.get<CreateAdService>(CreateAdService);
  });

  it('should be defined', () => {
    expect(createAdService).toBeDefined();
  });

  describe('execution', () => {
    const createAdCommand = new CreateAdCommand(createAdProps);
    it('should create a new ad', async () => {
      AdEntity.create = jest.fn().mockReturnValue({
        id: '047a6ecf-23d4-4d3e-877c-3225d560a8da',
      });
      const result: AggregateID = await createAdService.execute(
        createAdCommand,
      );
      expect(result).toBe('047a6ecf-23d4-4d3e-877c-3225d560a8da');
    });
    it('should throw an error if something bad happens', async () => {
      AdEntity.create = jest.fn().mockReturnValue({
        id: '047a6ecf-23d4-4d3e-877c-3225d560a8da',
      });
      await expect(
        createAdService.execute(createAdCommand),
      ).rejects.toBeInstanceOf(Error);
    });
    it('should throw an exception if Ad already exists', async () => {
      AdEntity.create = jest.fn().mockReturnValue({
        id: '047a6ecf-23d4-4d3e-877c-3225d560a8da',
      });
      await expect(
        createAdService.execute(createAdCommand),
      ).rejects.toBeInstanceOf(AdAlreadyExistsException);
    });
  });
});
