import { AggregateRoot, AggregateID } from '@mobicoop/ddd-library';
import {
  CreateRouteProps,
  Path,
  Role,
  RouteProps,
  PathType,
  Direction,
} from './route.types';
import { WaypointProps } from './value-objects/waypoint.value-object';
import { Route } from '../application/types/route.type';
import { v4 } from 'uuid';
import { RouteNotFoundException } from './route.errors';

export class RouteEntity extends AggregateRoot<RouteProps> {
  protected readonly _id: AggregateID;

  static create = async (create: CreateRouteProps): Promise<RouteEntity> => {
    let directions: Direction[];
    try {
      directions = await create.georouter.routes(
        this.getPaths(create.roles, create.waypoints),
        create.georouterSettings,
      );
      if (!directions || directions.length == 0)
        throw new RouteNotFoundException();
    } catch (e: any) {
      throw e;
    }
    let driverRoute: Route;
    let passengerRoute: Route;
    if (directions.some((route: Route) => route.type == PathType.GENERIC)) {
      driverRoute = passengerRoute = directions.find(
        (route: Route) => route.type == PathType.GENERIC,
      );
    } else {
      driverRoute = directions.some(
        (route: Route) => route.type == PathType.DRIVER,
      )
        ? directions.find((route: Route) => route.type == PathType.DRIVER)
        : undefined;
      passengerRoute = directions.some(
        (route: Route) => route.type == PathType.PASSENGER,
      )
        ? directions.find((route: Route) => route.type == PathType.PASSENGER)
        : undefined;
    }
    const routeProps: RouteProps = {
      driverDistance: driverRoute?.distance,
      driverDuration: driverRoute?.duration,
      passengerDistance: passengerRoute?.distance,
      passengerDuration: passengerRoute?.duration,
      fwdAzimuth: driverRoute
        ? driverRoute.fwdAzimuth
        : passengerRoute.fwdAzimuth,
      backAzimuth: driverRoute
        ? driverRoute.backAzimuth
        : passengerRoute.backAzimuth,
      distanceAzimuth: driverRoute
        ? driverRoute.distanceAzimuth
        : passengerRoute.distanceAzimuth,
      waypoints: create.waypoints,
      spacetimePoints: driverRoute
        ? driverRoute.spacetimePoints
        : passengerRoute.spacetimePoints,
    };
    return new RouteEntity({
      id: v4(),
      props: routeProps,
    });
  };

  validate(): void {
    // entity business rules validation to protect it's invariant before saving entity to a database
  }

  private static getPaths = (
    roles: Role[],
    waypoints: WaypointProps[],
  ): Path[] => {
    const paths: Path[] = [];
    if (roles.includes(Role.DRIVER) && roles.includes(Role.PASSENGER)) {
      if (waypoints.length == 2) {
        // 2 points => same route for driver and passenger
        paths.push(this.createGenericPath(waypoints));
      } else {
        paths.push(
          this.createDriverPath(waypoints),
          this.createPassengerPath(waypoints),
        );
      }
    } else if (roles.includes(Role.DRIVER)) {
      paths.push(this.createDriverPath(waypoints));
    } else if (roles.includes(Role.PASSENGER)) {
      paths.push(this.createPassengerPath(waypoints));
    }
    return paths;
  };

  private static createGenericPath = (waypoints: WaypointProps[]): Path =>
    this.createPath(waypoints, PathType.GENERIC);

  private static createDriverPath = (waypoints: WaypointProps[]): Path =>
    this.createPath(waypoints, PathType.DRIVER);

  private static createPassengerPath = (waypoints: WaypointProps[]): Path =>
    this.createPath(
      [waypoints[0], waypoints[waypoints.length - 1]],
      PathType.PASSENGER,
    );

  private static createPath = (
    points: WaypointProps[],
    type: PathType,
  ): Path => ({
    type,
    points,
  });
}

// import { IGeodesic } from '../interfaces/geodesic.interface';
// import { Point } from '../types/point.type';
// import { SpacetimePoint } from './spacetime-point';

// export class Route {
//   distance: number;
//   duration: number;
//   fwdAzimuth: number;
//   backAzimuth: number;
//   distanceAzimuth: number;
//   points: Point[];
//   spacetimePoints: SpacetimePoint[];
//   private geodesic: IGeodesic;

//   constructor(geodesic: IGeodesic) {
//     this.distance = undefined;
//     this.duration = undefined;
//     this.fwdAzimuth = undefined;
//     this.backAzimuth = undefined;
//     this.distanceAzimuth = undefined;
//     this.points = [];
//     this.spacetimePoints = [];
//     this.geodesic = geodesic;
//   }

//   setPoints = (points: Point[]): void => {
//     this.points = points;
//     this.setAzimuth(points);
//   };

//   setSpacetimePoints = (spacetimePoints: SpacetimePoint[]): void => {
//     this.spacetimePoints = spacetimePoints;
//   };

//   protected setAzimuth = (points: Point[]): void => {
//     const inverse = this.geodesic.inverse(
//       points[0].lon,
//       points[0].lat,
//       points[points.length - 1].lon,
//       points[points.length - 1].lat,
//     );
//     this.fwdAzimuth =
//       inverse.azimuth >= 0 ? inverse.azimuth : 360 - Math.abs(inverse.azimuth);
//     this.backAzimuth =
//       this.fwdAzimuth > 180 ? this.fwdAzimuth - 180 : this.fwdAzimuth + 180;
//     this.distanceAzimuth = inverse.distance;
//   };
// }
